#!/usr/bin/env python3

import math, itertools, sys, time
import SAT_lib, my_utils

SIZE=10

s=SAT_lib.SAT_lib()

a=[[s.alloc_BV(SIZE) for c in range(SIZE)] for r in range(SIZE)]
b=[[s.alloc_BV(SIZE) for c in range(SIZE)] for r in range(SIZE)]

def make_distinct_one_hots(s, lst):
    [s.OR_always(list(t)) for t in zip(*lst)]
    # also can be implemented as:
    #s.fix_BV_all_bits_1(s.BV_OR_list(lst))

def add_constraints(s, ar):
    for r in range(SIZE):
        for c in range(SIZE):
            s.make_one_hot(ar[r][c])

    # all numbers in all rows must be distinct:
    for r in range(SIZE):
        make_distinct_one_hots(s, [ar[r][c] for c in range(SIZE)])

    # ... in all columns as well:
    for c in range(SIZE):
        make_distinct_one_hots(s, [ar[r][c] for r in range(SIZE)])

add_constraints(s, a)
add_constraints(s, b)

t=[]
for r in range(SIZE):
    for c in range(SIZE):
        t.append(a[r][c]+b[r][c])

s.make_distinct_BVs(t)

def get_square(s,a):
    rt=[]
    for r in range(SIZE):
        l=[]
        for c in range(SIZE):
            l.append(int(math.log(SAT_lib.BV_to_number(s.get_BV_from_solution(a[r][c])), 2)))
        rt.append(l)
    return rt

def print_square(s,a):
    for r in range(SIZE):
        for c in range(SIZE):
            x=get_square(s,a)[r][c]
            print ("%d " % x, end='')
        print ("")

def print_MOLS(s, a, b):
    all_vals=[]
    for r in range(SIZE):
        for c in range(SIZE):
            x=get_square(s, a)[r][c]*10 + get_square(s,b)[r][c]
            print ("%02d " % x, end='')
            all_vals.append(x)
        print ("")

    assert len(set(all_vals))==SIZE*SIZE

# Knuth a
first=[
"3145926870",
"2819763504",
"9452307168",
"6208451793",
"8364095217",
"5981274036",
"4627530981",
"0576148329",
"1730689452",
"7093812645"]

for r in range(SIZE):
    for c in range(SIZE):
        tmp=SAT_lib.n_to_BV(1 << int(first[r][c]), SIZE)
        s.fix_BV(a[r][c], tmp)

# setting first column of mate (b) to [0..9]
for r in range(SIZE):
    c=0
    tmp=SAT_lib.n_to_BV(1 << r, SIZE)
    s.fix_BV(b[r][c], tmp)

sol=0

start_t=time.time()
print ("time at start: ", start_t)

if s.solve():
    while True:
        sol=sol+1
        print ("*** solution %d, seconds from start %d" % (sol, time.time()-start_t))
        print ("first:")
        print_square (s, a)
        print ("")
        print ("mate:")
        print_square (s, b)
        print ("")
        print ("concatenated:")
        print_MOLS(s, a, b)
        print ("")
        sys.stdout.flush()
        if s.fetch_next_solution()==False:
            break

print ("solutions: %d" % sol)

