#!/usr/bin/env python3

import math, itertools, sys, time, os
import SAT_lib, my_utils

ONLY_ONE_SOLUTION=True

SIZE=int(sys.argv[1])

_seed=int.from_bytes(os.urandom(4), byteorder='big')&0x7fffffff
print ("SIZE=", SIZE)
print ("SEED=", _seed)

s=SAT_lib.SAT_lib(seed=_seed)

a=[[s.alloc_BV(SIZE) for c in range(SIZE)] for r in range(SIZE)]
b=[[s.alloc_BV(SIZE) for c in range(SIZE)] for r in range(SIZE)]

def make_distinct_one_hots(s, lst):
    [s.OR_always(list(t)) for t in zip(*lst)]
    # also can be implemented as:
    #s.fix_BV_all_bits_1(s.BV_OR_list(lst))

def add_constraints(s, ar):
    for r in range(SIZE):
        for c in range(SIZE):
            s.make_one_hot(ar[r][c])

    # all numbers in all rows must be distinct:
    for r in range(SIZE):
        make_distinct_one_hots(s, [ar[r][c] for c in range(SIZE)])

    # ... in all columns as well:
    for c in range(SIZE):
        make_distinct_one_hots(s, [ar[r][c] for r in range(SIZE)])

add_constraints(s, a)
add_constraints(s, b)

t=[]
for r in range(SIZE):
    for c in range(SIZE):
        t.append(a[r][c]+b[r][c])

s.make_distinct_BVs(t)

# setting all columns to [0..9]
for _r in range(SIZE):
    _c=0
    tmp=SAT_lib.n_to_BV(1 << _r, SIZE)
    s.fix_BV(a[_r][_c], tmp)
    s.fix_BV(b[_r][_c], tmp)

# setting first row of $a$ to [0..9]
for _c in range(SIZE):
    _r=0
    tmp=SAT_lib.n_to_BV(1 << _c, SIZE)
    s.fix_BV(a[_r][_c], tmp)

def get_square(s,a):
    rt=[]
    for r in range(SIZE):
        l=[]
        for c in range(SIZE):
            l.append(int(math.log(SAT_lib.BV_to_number(s.get_BV_from_solution(a[r][c])), 2)))
        rt.append(l)
    return rt

def dump_square(s,a,name):
    rt=[]
    rt.append (name+":")
    for r in range(SIZE):
        t=""
        for c in range(SIZE):
            x=get_square(s,a)[r][c]
            t=t+("%d " % x)
        rt.append(t)
    return rt

def print_MOLS(s, a, b):
    all_vals=[]
    for r in range(SIZE):
        for c in range(SIZE):
            x=get_square(s, a)[r][c]*10 + get_square(s,b)[r][c]
            print ("%02d " % x, end='')
            all_vals.append(x)
        print ("")

    assert len(set(all_vals))==SIZE*SIZE

sol=0

start_t=int(time.time())
print ("time at start: ", start_t)

def pad_list_of_strings_by_max_string (l):
    return list(map(lambda s: s.ljust(max(map(len, l))), l))

def concat_list_of_strings_side_by_side(lst):
    rt=[]
    lists_total=len(lst)
    for i in range(len(lst[0])):
        s=""
        for j in range(lists_total):
            s=s+lst[j][i]
        rt.append(s)
    return rt

if s.solve():
    while True:
        sol=sol+1
        print ("*** solution %d, seconds from start %d" % (sol, int(time.time())-start_t))
        _a=dump_square (s, a,"first")
        a_padded=pad_list_of_strings_by_max_string (_a)
        _b=dump_square (s, b, "mate")
        b_padded=pad_list_of_strings_by_max_string (_b)
        print ("")
        for l in concat_list_of_strings_side_by_side([a_padded, [" | "]*len(a_padded), b_padded]):
            print (l)
        print ("")
        print ("concatenated:")
        print_MOLS(s, a, b)
        print ("")
        sys.stdout.flush()
        if ONLY_ONE_SOLUTION:
            exit(0)
        if s.fetch_next_solution()==False:
            break

print ("solutions: %d" % sol)

